#include <stdio.h>
#include <sys/types.h>
#include <strings.h>
#include <fcntl.h>
#include <termios.h>
#include <unistd.h>

#define X10_ON	2
#define X10_OFF	3
#define X10_DIM	5
#define VERSION	"v1.0"

void X10msg(int fd,u_char *msg,int len)
{
  int i;
  static u_char sync[16]={0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
                          0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff};

  if(write(fd,sync,16) < 16) printf("Error sending sync\n");
  if(write(fd,msg,len) < len) printf("Error sending message\n");
/*  tcdrain(fd); */
  while(0) {
    char buf[2];
    read(fd,buf,1);
    printf("%x\n",buf[0]);
  }
}

/* Convert 'abcdefghijklmnop' to house code for X10, EG: 'a' -> 0x60 */
u_char X10housecode(char c)
{
  static char *housecodes = "meckogainfdlphbj";

  c=tolower(c);
  if(c>='a' && c <= 'p')
    return ((strlen(housecodes) - strlen(index(housecodes,c)))<<4);
}

/* Set the base housecode for the unit */
void X10base_housecode(int fd,u_char housecode)
{
  static char hc[2]={0,0},answer[2];

  hc[1]=X10housecode(housecode);
  printf("Resetting your base housecode will erase any memory in your "
         "X10 control unit.\n\n"
         "Are you sure you want to set the base housecode on your"
         "control unit to %c? ",housecode);
  scanf("%1s",answer);
  if(tolower(answer[0])=='y') {
    X10msg(fd,hc,2);
    while(1) {
      char buf[1];

      read(fd,buf,1);
      printf("%d",buf[0]);
    }
  }
  else printf("Operation aborted!\n");
}

/* Return sum of len bytes */
u_char X10checksum(u_char *bytes,int len)
{
  int i;
  u_char sum;

  for(i=0;i<len;i++)
    sum+=bytes[i];

  return sum;
}

/* convert string of numbers in the form of "1,2,3,4" to a bitmap appropriate
   for the X10 control module */
u_short strtobm(char *string)
{
#define ITOBM(i)		(((i)<=16&&(i)>0)?(1<<(16-(i))):(0))
  int i;
  u_short rval=0,cur=0;

  if(string) {
    for(i=0;string[i];i++) {
      if(isdigit(string[i])) cur = cur*10 + string[i]-'0';
      else {
        rval |= ITOBM(cur);
        cur = 0;
      }
    }
    rval |= ITOBM(cur);
  }
  return rval;
}

X10direct(int fd,u_char command,char hc,char *str)
{
#define X10_MKBM_1_8(a)		(((a)&0xff00)>>8)
#define X10_MKBM_9_16(a)	(((a)&0xff))

  static u_char dc[6];
  u_short bm;

  bm = strtobm(str);

  dc[0] = 1; /* Direct command */
  dc[1] = command;
  dc[2] = X10housecode(hc);
  dc[3] = X10_MKBM_9_16(bm);
  dc[4] = X10_MKBM_1_8(bm);
  dc[5] = X10checksum(dc+1,4);

  X10msg(fd,dc,6);
}

void config_serial(int fd)
{
  static struct termios tios;

  tcgetattr(fd,&tios);
  tios.c_cflag = CLOCAL | CREAD | CS8 | B600;
  cfsetospeed(&tios,B600);
  cfsetispeed(&tios,B600);
  tcsetattr(fd,TCSADRAIN,&tios);
}

void usage(char *name)
{
  printf("x10 "VERSION" control program "
         "by Aaron Hightower (aaron@dseg.ti.com)\n"
         "\nUsage:\n\n"
         "%s: [-v] [-c housecode] [-n list] [-f list] [-d dimlevel,list]\n\n"
         "-v : verbose\n"
         "-c [a-p] : use alternate house code (default a)\n"
         "-n list : turn oN devices in list\n"
         "-f list : turn oFf devices in list\n"
         "-d dimlevel,list : dim devices in list to dimlevel\n"
         "list is a comma separated list of devices, each ranging from 1 to 16"
         "\ndimlevel is an integer from 0 to 15 (0 brightest)\n"
         "\nExamples:\n"
         "\tTurn on devices a1 a2 a3\n"
         "\t%s -n 1,2,3\n\n"
         "\tTurn off devices b1 b2 b3\n"
         "\t%s -c b -f 4,6\n\n"
         "\tdim lights 2 and 3 to 5th brightest setting\n"
         "\t%s -d 5,2,3\n\n"
          ,name,name,name,name,name);
  exit(0);
}

void main(int argc,char * const argv[])
{
  int i,fd,c;
  char dimlevel=0,verbose=0,housecode='a';
  char *dim=NULL,*portname=NULL,*optstr="z:d:vhc:n:f:";

  if(argc<=1) {usage(argv[0]);exit(0);}

  if(!portname) portname = (char *)getenv("X10_PORTNAME");
  if(!portname) portname = "/dev/x10";

  if((fd = open(portname,O_RDWR|O_NOCTTY,0)) == -1) {
    printf("Could not access serial port %s\nPlease set environment variable "
           "X10_PORTNAME to proper serial device\n"
           "or else make the appropriate soft link as superuser.\n",portname);
    exit(-1);
  }
  config_serial(fd);

  while((c = getopt(argc,argv,optstr)) != -1) {
    switch(c) {
      case 'z':
        X10base_housecode(fd,optarg?(*optarg):'a');
        break;
      case 'v':
        verbose=1;
        break;
      case 'c':
        housecode = optarg?(*optarg):'a';
        break;
      case 'd':
        {
	  if(optarg) { /* First number is the dimlevel (0..15) */
	    dim = strdup(optarg);
	    sscanf(optarg,"%d,%s",&dimlevel,dim);
            if(verbose)
              printf("Dimming %c-\"%s\" to %d\n",housecode,dim,dimlevel);
            X10direct(fd,X10_DIM + ((dimlevel & 0xf) << 4),housecode,dim);
	  }
	}
	break;
      case 'n':
        if(verbose) printf("Turning on %c-\"%s\"\n",housecode,optarg);
	X10direct(fd,X10_ON,housecode,optarg);
        break;
      case 'f':
        if(verbose) printf("Turning off %c-\"%s\"\n",housecode,optarg);
        X10direct(fd,X10_OFF,housecode,optarg);
        break;
      case 'h':
      case '?':
        usage(argv[0]);
        break;
    }
  }
  close(fd);
}
